import numpy as np

from .citrees import get_feature_importance
from .tree import get_nodes
from .utils import scientific_notation


def _rgb_to_hex(rgb):
    """Convert RGB to hex for colors.

    Parameters
    ----------
    rgb : 1-d like array
        RGB array of size 3.

    Returns
    -------
    hex_repr : str
        Hexidecimal representation of rgb.
    """

    hex_repr = "#{:02x}{:02x}{:02x}".format(rgb[0], rgb[1], rgb[2])
    return hex_repr

def _color_brew(n, alpha=1.0):
    """Generate n colors with equally spaced hues.

    Parameters
    ----------
    n : int
        The number of colors required.

    alpha : float
        Factor to upscale all the colors.

    Returns
    -------
    color_array : 2d array-like
        Arrat of size (n, 3), where each row is of form (R, G, B), which are 
        components of each color.
    """

    color_array = []

    # initialize saturation & value; calculate chroma & value shift
    s, v = 0.75, 0.9
    c = s * v
    m = v - c

    for h in np.arange(25, 385, 360. / n).astype(int):
        # calculate some intermediate values
        h_bar = h / 60.
        x = c * (1 - abs((h_bar % 2) - 1))
        # initialize RGB with same hue & chroma as our color
        rgb = [(c, x, 0),
               (x, c, 0),
               (0, c, x),
               (0, x, c),
               (x, 0, c),
               (c, 0, x),
               (c, x, 0)]
        r, g, b = rgb[int(h_bar)]

        # shift the initial RGB values to match value and store
        rgb = [(int(255 * (r + m))),
               (int(255 * (g + m))),
               (int(255 * (b + m)))]

        # perform upscaling
        rgb = np.array(rgb) * alpha
        np.putmask(rgb, rgb > 255, 255)

        color_array.append(rgb.astype(int))

    return np.array(color_array)


class GraphvizTreeExporter(object):
    """Export the tree using graphviz.

    Parameters
    ----------
    qnet : Qnet
        A Qnet instance

    outfile : str
        Output file to save results to

    response_name : str
        Name of the y variable that we are predicting

    feature_names : list
        Names of each of the features

    text_color : str
        Color to set the text

    edge_color : str
        Color to set the edges

    pen_width : int
        Width of pen for drawing boundaries
 
    dpi : int
        Image resolution
        
    rotate : bool
        If True, rotate the tree

    add_legend : bool
        If True, add a legend to the tree

    detailed_output : bool
        If False output probability of max likelihood of label in leaves (default), else output prob. distribution. 

    edge_font_color : str
        Color of edge label text

    min_size : int
        Minimum number of nodes to draw the tree

    labels : list
        List of all labels, optional

    Returns
    -------
    None
    """

    def __init__(
        self, 
        tree,
        outfile,
        response_name,
        feature_names, 
        text_color='black', 
        edge_color='gray',
        font_size=10,
        edge_label_color='deepskyblue4',
        pen_width=2,
        background_color='transparent',
        dpi=200,
        edge_fontcolor='grey14',
        rotate=False,
        add_legend=True,
        min_size=1,
        color_alpha=-1.5,
        labels = None,
        detailed_output=False):
        
        self.tree = tree
        self._total_samples = sum(tree.root.label_frequency.values())
        self.outfile = outfile
        self.response_name = response_name
        self.feature_names = feature_names
        self.text_color = text_color
        self.edge_color = edge_color
        self.font_size = font_size
        self.edge_label_color = edge_label_color
        self.pen_width = pen_width
        self.background_color = background_color
        self.dpi = dpi
        self.rotate = rotate
        self.add_legend = add_legend
        self.detailed_output = detailed_output
        self.edge_fontcolor = edge_fontcolor
        self.min_size = min_size
        self.color_alpha = color_alpha
        self.labels = labels
        
    def export(self):

        labels=self.labels
        if labels is None:
            labels = self.tree.labels_

        if not np.all([x in labels for x in self.tree.labels_]):
            raise ValueError('all tree labels not in specified lebelset')

        self.num_labels = len(labels)

        # esablish the color scheme
        color_brew = _color_brew(self.num_labels, alpha=self.color_alpha)
        self.class_colors = {labels[i]: color_brew[i] for i in range(self.num_labels)}

        nodes = get_nodes(
            self.tree.root, 
            get_leaves=True, 
            get_non_leaves=True)

        if len(nodes) < self.min_size:
            return False
        
        for i, node in enumerate(nodes):
            node.unique_id = i

        with open(self.outfile, 'w+') as f:
            self._write_tree_attributes(f)
            self._write_node_attributes(f)
            self._write_edge_attributes(f)

            leaf_node_ids = []
            final_string = "{rank = same;" + ';'.join(leaf_node_ids) + "}" 
            
            if self.add_legend:
                final_string += "{rank = same; LEGEND;1;}"

            final_string += '}'

            f.write(final_string)

        for i, node in enumerate(nodes):
            delattr(node, 'unique_id')

        return True

    def _write_tree_attributes(self, f):
        f.write('graph Tree {\n')
        f.write('node [shape=box')

        style = ['filled', 'rounded']
        style = ', '.join(style)

        f.write(''', style="{}",color="{}",penwidth="{}",fontcolor="{}", \
                fontname=helvetica] ;\n'''.format(style,
                                                  self.text_color,
                                                  self.pen_width,
                                                  self.text_color))

        f.write('''graph [ranksep="0 equally", splines=straight, \
                bgcolor={}, dpi={}] ;\n'''.format(self.background_color, self.dpi))

        f.write('edge [fontname=helvetica, fontweight=bold,fontcolor={},color={}] ;\n'.format(self.edge_fontcolor,
            self.edge_color))

        if self.rotate:
            f.write('rankdir=LR ;\n')

        if self.add_legend:
            self._write_legend(f)


    def _write_legend(self, f):
        label = ''

        if self.response_name is not None:
            label += "Response: {}\n".format(self.response_name)

        label += "Classes: {}\n".format(' | '.join(self.tree.labels_))

        f.write('LEGEND [label="{}",shape=note,align=left,\
                style=filled,fillcolor="slategray",\
                fontcolor="white"];\n'.format(label))

    def _find_node_color(self, node):
        label_freq = node.label_frequency
        total = float(sum(label_freq.values()))
        coloring = np.zeros(3)
        for val, freq in label_freq.items():
            coloring += freq * self.class_colors[val]

        coloring = (coloring / total).astype(int)

        return _rgb_to_hex(coloring)

    def _write_node_attribute(self, f, node):

        node_id = node.unique_id

        node_labels = ''

        if (node.left is None) and (node.right is None):
            max_index = np.argmax(node.value)
            max_val = round(node.value[max_index], 3)
            occurence = sum(node.label_frequency.values()) / self._total_samples
            occurence = round(occurence, 3)
            prediction = self.tree.labels_[max_index]
            
            if not self.detailed_output:            
                node_labels += '{}\nProb: {}\nFrac: {}'.format(
                    prediction, 
                    max_val,
                    occurence)
            else:
                node_labels += '{}\nProb: {}\nFrac: {}'.format(
                    prediction,
                    ' '.join([l+':'+"{:1.1f}".format(v)
                              for (l,v) in zip(self.tree.labels_,node.value)]),
                    occurence)
        else:
            if self.feature_names is not None:
                node_labels += self.feature_names[node.col]
            pval = scientific_notation(node.col_pval)
            # node_labels += " pval: {}".format(pval)

        node_color = self._find_node_color(node)
        f.write('{} [label="{}"'.format(node_id, node_labels))
        f.write(', fillcolor="{}"] ;\n'.format(
            node_color))

    def _write_node_attributes(self, f):
        nodes = get_nodes(
            self.tree.root, 
            get_leaves=True, 
            get_non_leaves=True)

        for node in nodes:
            self._write_node_attribute(f, node)

    def _write_edge_labels(self, f, parent_node, node, node_type):

        string_format = '{} -- {} [label="{}",penwidth={}] ;\n'

        parent_id = parent_node.unique_id

        if node_type == 'left':
            threshold = parent_node.lthreshold
        elif node_type == 'right':
            threshold = parent_node.rthreshold
        else:
            raise ValueError

        child_id = node.unique_id
        edge_label = ' ' + '\\n '.join(threshold)
        
        f.write(string_format.format(parent_id,
                                     child_id,
                                     edge_label,
                                     self.pen_width))

    def _write_edge_attribute(self, f, node):
        left_node = node.left
        right_node = node.right

        if left_node is not None:
            self._write_edge_labels(f, node, left_node, 'left')
        if right_node is not None:
            self._write_edge_labels(f, node, right_node, 'right')

    def _write_edge_attributes(self, f):
        nodes = get_nodes(
            self.tree.root, 
            get_leaves=True, 
            get_non_leaves=True)

        for node in nodes:
            self._write_edge_attribute(f, node)



class QnetGraphExporter(object):
    """Export the qnet as a graph to a dot file format.

    Parameters
    ----------
    qnet : Qnet
        A Qnet instance

    outfile : str
        Output file to save results to

    threshold : float
        Numeric cutoff for edge weights. If the edge weights exceed 
        this cutoff, then we include it into the graph.

    Returns
    -------
    None
    """

    def __init__(
        self, 
        qnet,
        outfile,
        threshold):

        self.qnet = qnet
        self.outfile = outfile
        self.threshold = threshold

        # `nodes` will store the column as indices
        self.nodes = set()

        # `edges` will be a list of dictionaries.
        self.edges = []

    def _col_to_desc(self, col):
        """Get the description of `col`. """
        return self.qnet.feature_names[col]

    def _get_nodes_and_edges(self):

        for i, tree in self.qnet.estimators_.items():
            feature_imp = get_feature_importance(tree, normalize=True)
            for col, imp in feature_imp.items():
                if imp >= self.threshold:
                    self.nodes.add(i)
                    self.nodes.add(col)

                    edge = {
                        'start': col,
                        'end': i,
                        'weight': imp
                    }
                    self.edges.append(edge)


    def _write_start(self, f):
        """Write the starting text of the .dot file. """
        f.write('strict digraph  {\n')

    def _write_nodes(self, f):
        """Write the nodes to the .dot file. """

        for node in self.nodes:
            f.write('{};\n'.format(self._col_to_desc(node)))

    def _write_edges(self, f):
        """Write the edges to the .dot file. """

        fmt = '{} -> {}  [weight="{}"];\n'
        for edge in self.edges:
            start = self._col_to_desc(edge['start'])
            end = self._col_to_desc(edge['end'])
            f.write(fmt.format(
                start,
                end,
                edge['weight']))

    def _write_end(self, f):
        """Write the ending text of the .dot file. """
        f.write('}')

    def export(self):
        self._get_nodes_and_edges()

        with open(self.outfile, 'w+') as f:
            self._write_start(f)
            self._write_nodes(f)
            self._write_edges(f)
            self._write_end(f)



import unittest
from unittest import TestCase

import pandas as pd
import numpy as np

from quasinet.qnet import Qnet, qdistance, membership_degree, qdistance_matrix

class TestQnet( TestCase):
    DATA_DIR = 'examples/example_data/'

    def load_data(self):
        if not hasattr(self, 'X'):
            max_rows = 5
            X = pd.read_csv(self.DATA_DIR + 'cchfl_test.csv')
            self.features = list(X.columns.astype(str))[:max_rows]
            X = X.values.astype(str)[:, :max_rows]
            self.X = X

        return self.X

    def test_create_qnet(self):
        if not hasattr(self, 'myqnet'):
            X = self.load_data()
            myqnet = Qnet(feature_names=self.features, n_jobs=1)
            myqnet.fit(X)
            self.myqnet = myqnet

        return self.myqnet

    def test_qdistance(self):
        X = self.load_data()

        seq1 = X[1]
        seq2 = X[2]
        myqnet = self.test_create_qnet()
        qdist = qdistance(seq1, seq2, myqnet, myqnet) 
        self.assertGreaterEqual(qdist, 0)

    @staticmethod
    def compute_qdistance_matrix(seqs1, seqs2, myqnet1, myqnet2):
        """Compute the qdistance matrix without any optimizations."""
        qdist_matrix = np.empty((len(seqs1), len(seqs2)))
        for i, seq1 in enumerate(seqs1):
            for j, seq2 in enumerate(seqs2):
                qdist_matrix[i, j] = qdistance(seq1, seq2, myqnet1, myqnet2)

        return qdist_matrix

    def test_qdistance_matrix(self):
        X = self.load_data()
        qnet1 = self.test_create_qnet()

        qnet2 = self.test_create_qnet()

        seqs1 = X[:4]
        seqs2 = X[4:7]

        # test when the seqs are different and qnets are different
        expected_qdist_matrix = TestQnet.compute_qdistance_matrix(
            seqs1,
            seqs2,
            qnet1,
            qnet2)
        actual_qdist_matrix = qdistance_matrix(seqs1, seqs2, qnet1, qnet2)
        self.assertTrue(np.allclose(expected_qdist_matrix, actual_qdist_matrix))

        # test when the seqs are the same and qnet is the same
        expected_qdist_matrix = TestQnet.compute_qdistance_matrix(
            seqs1,
            seqs1,
            qnet1,
            qnet1)
        actual_qdist_matrix = qdistance_matrix(seqs1, seqs1, qnet1, qnet1)
        self.assertTrue(np.allclose(expected_qdist_matrix, actual_qdist_matrix))

    


    def test_membership_degree(self):
        X = self.load_data()
        seq1 = X[1]
        myqnet = self.test_create_qnet()
        mem_degree = membership_degree(seq1, myqnet)


if __name__ == "__main__":
    unittest.main()



import numpy as np
from joblib import dump, load, delayed, Parallel
import numba
# from numba import njit, prange
# from numba.core import types

from .citrees import CITreeClassifier
from .metrics import js_divergence
from .tree import Node, get_nodes
from .utils import assert_array_rank, assert_string_type
from .export import GraphvizTreeExporter, QnetGraphExporter
from ._config import get_config

class Qnet(object):
    """Qnet architecture.

    Parameters
    ----------

    feature_names : list
        List of names describing the features

    min_samples_split : int
        Minimum samples required for a split

    alpha : float
        Threshold value for selecting feature with permutation tests. Smaller
        values correspond to shallower trees

    max_depth : int
        Maximum depth to grow tree

    max_feats : str or int
        Maximum feats to select at each split. String arguments include 'sqrt',
        'log', and 'all'

    early_stopping : bool
        Whether to implement early stopping during feature selection. If True,
        then as soon as the first permutation test returns a p-value less than
        alpha, this feature will be chosen as the splitting variable

    verbose : bool or int
        Controls verbosity of training and testing

    random_state : int
        Sets seed for random number generator

    n_jobs : int
        Number of CPUs to use when training
    """

    def __init__(
        self, 
        feature_names,
        min_samples_split=2,
        alpha=.05,
        max_depth=-1,
        max_feats=-1,
        early_stopping=False,
        verbose=0,
        random_state=None,
        n_jobs=1):

        self.feature_names = feature_names
        self.min_samples_split = min_samples_split
        self.alpha = alpha
        self.max_depth = max_depth
        self.max_feats = max_feats
        self.early_stopping = early_stopping
        self.verbose = verbose
        self.random_state = random_state
        self.n_jobs = n_jobs
        self.mixed = False

    def __repr__(self):
        return "qnet.Qnet"

    def __str__(self):
        return self.__repr__()

    def _parallel_fit_tree(self, tree, X, col):
        new_features = np.copy(X)
        new_features[:, col] = get_config()['nan_value']
        tree.fit(new_features, X[:, col])

        return tree

    def _check_input_size(self, size):
        feature_size = len(self.feature_names)
        if size != feature_size:
            string = 'The number of input features ({}) ' + \
                     'must match size of `feature_names` ({})!'
            string = string.format(size, feature_size)
            raise ValueError(string)

    def _check_is_fitted(self):
        if not hasattr(self, 'estimators_'):
            raise ValueError('You need to call `fit` first! ')

    def fit(self, X, index_array=None):
        """Train the qnet

        Examples
        ----------
        >>> from quasinet import qnet
        >>> X = load_data()
        >>> myqnet = qnet.Qnet(feature_names=feature_names)
        >>> myqnet.fit(X)

        Parameters
        ----------
        X : 2d array-like
            Array of features

        index_array : 1d array-like
            Array of indices to generate estimators for. Uses all indices by default or if set to None.

        Returns
        -------
        self : Qnet
            Instance of Qnet class
        """

        assert_array_rank(X, 2)
        
        self._check_input_size(X.shape[1])

        assert_string_type(X, 'X')

        # Instantiate base tree models
        self.estimators_ = {}

        nan_value = get_config()['nan_value']
        if np.any(np.all(X == nan_value, axis=0)):
            raise ValueError('There is at least one column where all'
                + ' values are `{}`. Please remove those columns!'.format(nan_value))

        # TODO: we may not have any trees created. When that's the
        # case, we may want to predict an equal probability distribution

        trees = []
        for col in np.arange(0, X.shape[1]):
            tree = CITreeClassifier(
                min_samples_split=self.min_samples_split,
                alpha=self.alpha,
                selector='chi2',
                max_depth=self.max_depth,
                max_feats=self.max_feats,
                early_stopping=self.early_stopping,
                verbose=self.verbose,
                random_state=self.random_state)
            trees.append(tree)

        if index_array is None:
            index_array=range(0, X.shape[1])

        # changing to  multiprocessing from loky @ishanu June 2022
        trees = Parallel(n_jobs=self.n_jobs, backend='multiprocessing')(
            delayed(self._parallel_fit_tree)(
                trees[col], X, col)
            for col in index_array
            )

        for col, tree in enumerate(trees):
            self.estimators_[col] = tree

        return self


    def mix(self,qnet_2,feature_name_list):
        """Take column from qnet_2, and switch its estimator with current qnet in-place. Makes it possible to simulate behavior of current qnet if some of the estimators were behaving like a second one, and can be used to identify the maximally divergent rules of orgnaization between the two models. Also sets the attribute self.mixed to be True.

        Parameters
        ----------
        qnet_2 : Qnet
            A Qnet instance

        feature_name_list : list
            A list of variable (feature) names that would be replaced in self from qnet_2

        Returns
        -------
        None
        """

        #check if qnet_2 is a qnet
        if not hasattr(qnet_2, 'feature_names'):
            raise ValueError('qnet_2 must be a qnet! ')
        
        #check if columns match
        if len([x for x in qnet_2.feature_names if x not in self.feature_names])>0:
            print("some qnet_2 features not present in current qnet. cannot mix")
        #check if estimators_ exist for both qnets

        if not hasattr(self, 'estimators_'):
            raise ValueError('You need to call `fit` first! ')
        if not hasattr(qnet_2, 'estimators_'):
            raise ValueError('You need to call `fit` first on qnet_2! ')

        column_list=[list(self.feature_names).index(x) for x in feature_name_list]
        column_list2=[list(qnet_2.feature_names).index(x) for x in feature_name_list]

        # find column_list form feature_list
        for col,col2 in zip(column_list,column_list2):
            self.estimators_[col]=qnet_2.estimators_[col2]

        if hasattr(self, 'col_to_non_leaf_nodes'):
            del self.col_to_non_leaf_nodes
        self._map_col_to_non_leaf_nodes()
        self.mixed=True

    def _map_col_to_non_leaf_nodes(self):
        """Get a mapping from column indexes to an array of all 
        non-leaf node indexes from the corresponding tree.

        We will cache the results because this function is called
        multiple times and it doesn't depend on any input argument.

        Parameters
        ----------
        None

        Returns
        -------
        prob_distributions : dictionary
            dictionary mapping indexes to a list of non-leaf nodes.
        """

        if hasattr(self, 'col_to_non_leaf_nodes'):
            pass
        else:
            self._check_is_fitted()
            
            col_to_nodes = {}
            for col, clf_tree in self.estimators_.items():
                nodes = get_nodes(
                    clf_tree.root, 
                    get_leaves=False, get_non_leaves=True)
                
                node_cols = np.array([node.col for node in nodes])

                col_to_nodes[col] = np.unique(node_cols)
            self.col_to_non_leaf_nodes = col_to_nodes

        return self.col_to_non_leaf_nodes


    def clear_attributes(self):
        """Remove the unneeded attributes to save memory.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """

        self._check_is_fitted()

        if hasattr(self, 'col_to_non_leaf_nodes'):
            delattr(self, 'col_to_non_leaf_nodes')

        new_estimator = {}
        for col, tree in self.estimators_.items():
            new_tree = CITreeClassifier(min_samples_split=self.min_samples_split,
                                        alpha=self.alpha,
                                        selector='chi2',
                                        max_depth=self.max_depth,
                                        max_feats=self.max_feats,
                                        early_stopping=self.early_stopping,
                                        verbose=self.verbose,
                                        random_state=self.random_state)

            new_tree.root = tree.root
            new_tree.labels_ = tree.labels_
            new_estimator[col] = new_tree

        delattr(self, 'estimators_')

        self.estimators_ = new_estimator

    def _combine_distributions(self, distributions):
        """Given a list of distributions, combine them together into
        an array.

        Parameters
        ----------
        distributions : list
            list of dictionaries mapping responses to probabilities.

        Returns
        -------
        array : 2d array-like
            the resulting array from combining the distributions.

        all_responses : list
            list of the responses corresponding to the array
        """

        all_responses = []
        for distrib in distributions:
            all_responses.extend(distrib.keys())

        all_responses = list(set(all_responses))

        res_to_index = {res: i for i, res in enumerate(all_responses)}

        array = np.zeros((len(distributions), len(all_responses)))

        for i, distrib in enumerate(distributions):
            for k, v in distrib.items():

                array[i, res_to_index[k]] = v

        return array, all_responses

    def predict_distribution(self, column_to_item, column):
        """Predict the probability distribution for a given column.

        It may be the case that a certain column value does not appear in 
        the resulting output. If that happens, that means the probability
        of that column is 0.

        Parameters
        ----------
        column_to_item : dict
            dictionary mapping the column to the values the columns take

        column : int
            column index

        Returns
        -------
        output : dictionary
            dictionary mapping possible column values to probability values
        """

        self._check_is_fitted()

        root = self.estimators_[column].root

        if len(column_to_item) == 0:
            distributions = dict(root.label_frequency)
            values = float(sum(distributions.values()))
            distributions = {k: v / values for k, v in distributions.items()}
            return distributions

        nodes = get_nodes(root)
        distributions = []
        for node in nodes:
            if node.col in column_to_item:
                if column_to_item[node.col] in node.lthreshold:
                    next_node = node.left
                elif column_to_item[node.col] in node.rthreshold:
                    next_node = node.right
                else:
                    continue

                distributions.append(next_node.label_frequency)

        if len(distributions) == 0:
            distributions = [root.label_frequency]

        distributions, columns = self._combine_distributions(distributions)
        
        total_frequency = np.sum(distributions)        
        distributions = distributions * np.sum(distributions, axis=1)[:, None]
        distributions /= total_frequency

        distributions = np.mean(distributions, axis=0)
        distributions /= np.sum(distributions)

        output = {columns[i]: distributions[i] for i in range(len(columns))}
        return output


    def predict_distributions(self, seq):
        """Predict the probability distributions for all the columns.

        If you do not want to set a particular value for an index of `seq`, 
        then set the value at the index to the global `nan_value`. By default, 
        this value is the empty string.

        The length of the input sequence must match the size of `feature_names`.

        Parameters
        ----------
        seq : list
            list of values

        Returns
        -------
        prob_distributions : list
            list of dictionaries of probability distributions, one for each index 
        """

        self._check_is_fitted()
        self._check_input_size(len(seq))

        index_to_non_leaf_nodes = self._map_col_to_non_leaf_nodes()

        prob_distributions = [None] * (max(self.estimators_.keys()) + 1)
        for col in self.estimators_.keys():
            col_to_item = {i: seq[i] for i in index_to_non_leaf_nodes[col]}
            distrib = self.predict_distribution(col_to_item, col)
            prob_distributions[col] = distrib

        return prob_distributions

    def _predict_distributions_numba(self, seq):
        raise NotImplementedError
        # prob_distributions = self.predict_distributions(seq)

        # numba_prob_distributions = []
        # for prob_distribution in prob_distributions:
        #     d = numba.typed.Dict.empty(
        #         key_type=types.unicode_type,
        #         value_type=types.float32)

        #     for k, v in prob_distribution.items():
        #         d[k] = v
        #     numba_prob_distributions.append(d)

        # return numba_prob_distributions

    def _predict_proba(self, X):

        self._check_is_fitted()

        raise NotImplementedError


def _assert_is_qnet(object_):
    """Check that the object is a qnet

    Parameters
    ----------
    object_ : obj
        Any object to check whether it is a `Qnet` or not.

    Returns
    -------
    None
    """

    if not isinstance(object_, Qnet):
        raise ValueError('The input is not a qnet.')

def _combine_two_distribs(seq1_distrib, seq2_distrib):
    """Combine two distributions together.

    The two distributions may contain different random variables.
    If one random variable doesn't exist in one of the distributions,
    then we set the probability to 0.

    Parameters
    ----------
    seq1_distrib : dict
        Dictionary mapping random variables to probabilities.

    seq2_distrib : dict
        Dictionary mapping random variables to probabilities.

    Returns
    -------
    output : 2d array-like
        array of shape (2, total unique random variables)
    """

    # make sure both distribution contains the same responses
    for seq1_response in seq1_distrib.keys():
        if seq1_response not in seq2_distrib:
            seq2_distrib[seq1_response] = 0.0

    for seq2_response in seq2_distrib.keys():
        if seq2_response not in seq1_distrib:
            seq1_distrib[seq2_response] = 0.0

    num_responses = len(seq2_distrib)

    distrib = np.empty((2, num_responses))
    for i, x in enumerate(seq1_distrib.keys()):
        distrib[0, i] = seq1_distrib[x]
        distrib[1, i] = seq2_distrib[x]

    return distrib


def _qdistance_with_prob_distribs(distrib1, distrib2):
    """
    
    NOTE: using njit may worsen speed performance

    Parameters
    ----------
    distrib1 : list
        List of dictionaries. Each dictionary maps random variables 
        to probabilities.

    distrib2 : list
        List of dictionaries. Each dictionary maps random variables 
        to probabilities.

    Returns
    -------
    avg_divergence : scalar
        qdistance
    """

    total_divergence = 0
    total = 0
    for i, seq1_distrib in enumerate(distrib1):
            
        seq2_distrib = distrib2[i]

        if seq2_distrib is None or seq1_distrib is None:
            continue
        
        distrib = _combine_two_distribs(seq1_distrib, seq2_distrib)

        total_divergence += np.sqrt(js_divergence(distrib[0], distrib[1]))

        total += 1
        
    avg_divergence = total_divergence / total

    return avg_divergence


def qdistance(seq1, seq2, qnet1, qnet2):
    """Compute the Jensen-Shannon of discrete probability distributions.

    Parameters
    ----------
    seq1 : 1d array-like
        Array of values

    seq2 : 1d array-like
        Array of values

    qnet1 : Qnet
        the Qnet that `seq1` belongs to 

    qnet2 : Qnet
        the Qnet that `seq2` belongs to 

    Returns
    -------
    output : numeric
        qdistance
    """

    if not isinstance(seq1, np.ndarray) or not isinstance(seq2, np.ndarray):
        raise ValueError('You must pass in arrays as sequences.')

    if len(seq1.shape) != 1 or len(seq2.shape) != 1:
        raise ValueError('The two sequences must be 1d arrays.')

    #if seq1.shape[0] != seq2.shape[0]:
    #    raise ValueError('The two sequences must be of equal lengths.')

    seq1_distribs = qnet1.predict_distributions(seq1)
    seq2_distribs = qnet2.predict_distributions(seq2)

    if seq1.shape[0] != seq2.shape[0]:
        keys=seq1_distribs[0].keys() & seq2_distribs[0].keys()
        seq1_distribs=[{i:seq1_distribs[0][i] for i in keys}]
        seq2_distribs=[{i:seq2_distribs[0][i] for i in keys}]
    
    divergence = _qdistance_with_prob_distribs(seq1_distribs, seq2_distribs)
    return divergence

def membership_degree(seq, qnet):
    """Compute the membership degree of a sequence in a qnet.

    Parameters
    ----------
    seq : 1d array-like
        Array of values

    qnet : Qnet
        the Qnet that `seq` belongs to 

    Returns
    -------
    membership_degree : numeric
        membership degree
    """

    assert_array_rank(seq, 1)

    if len(seq) == 0:
        raise ValueError('The sequence cannot be empty.')
    
    seq_distribs = qnet.predict_distributions(seq)

    index_probs = []
    for index, c in enumerate(seq):
        if seq_distribs[index] is None:
            continue

        distrib = seq_distribs[index]
        if distrib is None:
            continue
        
        if c in distrib:
            index_prob = distrib[c]
            
            if index_prob != 0:
                index_probs.append(index_prob)

    membership_degree = np.sum(np.log(index_probs))

    return membership_degree

def _qdistance_matrix_with_distribs(seqs1_distribs, seqs2_distribs, symmetric):
    """Compute the qdistance matrix using the distributions.

    Parameters
    ----------
    seqs1_distribs : list
        List of lists. Each sublist is a list of dictionaries mapping 
        random variables to probabilities.

    seqs2_distribs : list
        Same as `seqs1_distribs`

    symmetric : bool
        whether the distance matrix is symmetric or not

    Returns
    -------
    distance_matrix : 2d array-like
        distance matrix
    """

    num_seqs1 = len(seqs1_distribs)
    num_seqs2 = len(seqs2_distribs) 

    distance_matrix = np.empty((num_seqs1, num_seqs2))
    for i in np.arange(num_seqs1):
    # for i in prange(num_seqs1):
        for j in np.arange(num_seqs2):

            if symmetric and (i >= j):
                dist = 0.0
            else:
                dist = _qdistance_with_prob_distribs(seqs1_distribs[i], 
                                                     seqs2_distribs[j])

            distance_matrix[i, j] = dist

    if symmetric: 
        distance_matrix = distance_matrix + distance_matrix.T
    return distance_matrix


def qdistance_matrix(seqs1, seqs2, qnet1, qnet2):
    """Compute a distance matrix with the qdistance metric.

    Parameters
    ----------
    seqs1: 2d array-like
        Array of values

    seqs2: 2d array-like
        Array of values

    qnet1 : Qnet
        the Qnet that `seqs1` belongs to 

    qnet2 : Qnet
        the Qnet that `seqs2` belongs to 

    Returns
    -------
    distance_matrix : 2d array-like
        distance matrix
    """

    assert_array_rank(seqs1, 2)
    assert_array_rank(seqs2, 2)

    if seqs1.shape[1] != seqs2.shape[1]:
        raise ValueError('The columns of the two matrices must be equal.')

    if seqs1.shape[0] != seqs2.shape[0]:
        symmetric = False
    else:
        symmetric = np.all(seqs1 == seqs2) and (qnet1 == qnet2)

    # WARNING: do not try to access seqs1_distribs with non-numba code, 
    # as it will create a segmentation fault
    # this is likely a bug due to Numba
    # seqs1_distribs = numba.typed.List()
    seqs1_distribs = []
    for seq1 in seqs1:
        seqs1_distribs.append(qnet1.predict_distributions(seq1))

    if symmetric:
        seqs2_distribs = seqs1_distribs
    else:
        # seqs2_distribs = numba.typed.List()
        seqs2_distribs = []
        for seq2 in seqs2:
            seqs2_distribs.append(qnet2.predict_distributions(seq2))

    # seqs1_distribs = [qnet1.predict_distributions_numba(seq) for seq in seqs1]
    # seqs2_distribs = [qnet2.predict_distributions_numba(seq) for seq in seqs2]

    distance_matrix = _qdistance_matrix_with_distribs(seqs1_distribs, 
                                                      seqs2_distribs,
                                                      symmetric=symmetric)

    return distance_matrix
    

def load_qnet(f):
    """Load the qnet from a file.

    Parameters
    ----------
    f : str
        File name.
 
    Returns
    -------
    qnet : Qnet
    """

    qnet = load(f)
    assert isinstance(qnet, Qnet)

    return qnet

def save_qnet(qnet, f, low_mem=False):
    """Save the qnet to a file.

    NOTE: The file name must end in `.joblib`

    TODO: using joblib is actually less memory efficient than using pickle.
    However, I don't know if this is a general problem or this only happens
    under certain circumstances.

    TODO: we may have to delete and garbage collection some attributes in the qnet
    to save memory. For example, `.feature_importances_`, `.available_features_`

    Parameters
    ----------
    qnet1 : Qnet
        A Qnet instance

    f : str
        File name

    low_mem : bool
        If True, save the Qnet with low memory by deleting all data attributes 
        except the tree structure
 
    Returns
    -------
    None
    """
    import dill as pickle
    assert isinstance(qnet, Qnet)
    if not f.endswith('.joblib'):
        raise ValueError('The outfile must end with `.joblib`')

    if low_mem:
        if not hasattr(qnet, 'mixed'):
            qnet.clear_attributes()
        else:
            if not qnet.mixed:
                qnet.clear_attributes()
    filehandler = open(f,"wb")    
    pickle.dump(qnet, filehandler) 


def export_qnet_tree(qnet, index,
                     outfile,
                     outformat='graphviz',
                     detailed_output=False,
                     pen_width=3,
                     edge_color='black',
                     edge_label_color='black',
                     dpi=200,
                     text_color='black',
                     font_size=10,
                     background_color='transparent',
                     rotate=False,
                     edge_fontcolor='grey14',
                     min_size = 1,
                     color_alpha = 1.5,
                     labels=None,
                     add_legend=False):
    """Export a tree from `qnet`. The `index` determines which tree to export. 

    Parameters
    ----------
    qnet : Qnet
        A Qnet instance

    index : int
        Index of the tree to export
 
    outformat : str
        Can only be `graphviz` for now. This will output a `.dot` file, which you 
        can then compile using a command like `dot -Tpng file.dot -o file.png`
    detailed_output : bool
        If True return detailed probabilities of output labels in leaf nodes. default: False
    text_color : str
        Color to set the text

    edge_color : str
        Color to set the edges

    pen_width : int
        Width of pen for drawing boundaries
 
    dpi : int
        Image resolution
        
    rotate : bool
        If True, rotate the tree

    add_legend : bool
        If True, add a legend to the tree

    edge_font_color : str
        Color of edge label text

    min_size : int
        Minimum number of nodes to draw the tree

    color_alpha : float
        Parameter for color brightness

    labels : list
        List of all labels, optional

    Returns
    -------
    None
    """

    if index not in qnet.estimators_:
        print('We do not have a tree for index {}! '.format(index))
        return 

    tree = qnet.estimators_[index]
    feature_names = qnet.feature_names

    if outformat == 'graphviz':
        exporter = GraphvizTreeExporter(
            tree=tree,
            outfile=outfile,
            response_name=feature_names[index],
            feature_names=feature_names,
            pen_width=pen_width,
            edge_color=edge_color,
            edge_label_color=edge_label_color,
            add_legend=add_legend,
            font_size = font_size,
            background_color = background_color,
            dpi = dpi,
            rotate = rotate,
            text_color = text_color,
            edge_fontcolor=edge_fontcolor,
            min_size=min_size,
            color_alpha=color_alpha,
            labels = labels,
            detailed_output=detailed_output)
        exporter.export()

    else:
        raise NotImplementedError


def export_qnet_graph(qnet, threshold, outfile):
    """Export the `qnet` as a graph of dependencies. The output will be in
    the `.dot` file format for graphs.

    Parameters
    ----------
    qnet : Qnet
        A Qnet instance
 
    threshold : float
        Numeric cutoff for edge weights. If the edge weights exceed 
        this cutoff, then we include it into the graph.

    outfile : str
        File name to save to.

    Returns
    -------
    None
    """

    exporter = QnetGraphExporter(
        qnet, 
        outfile=outfile,
        threshold=threshold)

    exporter.export()

    



import os
import glob

from Bio import SeqIO
import numpy as np

from . import qnet as qnet

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
TRAINED_QNET_DIR = os.path.join(BASE_DIR, 'qnet_trees/')

ALL_QNET_TYPES = ['coronavirus', 'influenza']
CORONA_OPTIONS = ['bat', 'rat', 'game', 'covid19']

INFLUENZA_PROTEINS = ['na', 'ha']
INFLUENZA_TYPES = ['h1n1', 'h3n2']

def _get_possible_years(basedir):
    """Given a base directory for influenza, get the possible years.

    Parameters
    ----------
    basedir : str
        Base directory to search

    Returns
    -------
    possible_years : list
        list of possible years
    """

    possible_years = []
    for f in glob.glob(basedir + "*.joblib"):
        f = os.path.basename(f)
        year = int(f.replace('.joblib', '').split('_')[0])
        possible_years.append(year)

    possible_years.sort()

    return possible_years


def list_trained_qnets():
    """List the possible qnets we can use.

    Parameters
    ----------
    None

    Returns
    -------
    None
    """

    print('Possible qnets: \n')
    print('`coronavirus` options:')
    print('\t' + '\n\t'.join(CORONA_OPTIONS))
    
    print('\n`influenza` options:')
    influenza_options = []

    for protein in INFLUENZA_PROTEINS:
        for type_ in INFLUENZA_TYPES:
            base_dir = os.path.join(
            TRAINED_QNET_DIR, 
            'influenza', 
            '{}human{}/'.format(type_, protein.upper()))

            possible_years = _get_possible_years(base_dir)
            for year in possible_years:
                influenza_options.append('{};{};{}'.format(type_, protein, year))

    print('\t' + '\n\t'.join(influenza_options))


def load_trained_qnet(qnet_type, extra_descriptions):
    """Load the pre-trained qnet.

    Examples
    ----------
    >>> load_qnet('coronavirus', 'bat')
    >>> load_qnet('influenza', 'h1n1;na;2009')

    Parameters
    ----------
    qnet_type : str
        The type of qnet to load

    extra_descriptions : str
        Extra descriptions for which qnet to load. The descriptions must be
        split by `;` for influenza.

    Returns
    -------
    trained_qnet : qnet.Qnet
        A trained qnet
    """
    
    qnet_type = qnet_type.lower()
    extra_descriptions  = extra_descriptions.lower()
    
    if qnet_type not in ALL_QNET_TYPES:
        raise ValueError('`qnet_type`: {} is not in {}'.format(
            qnet_type, ALL_QNET_TYPES))
                         
    if qnet_type == 'coronavirus':
        if extra_descriptions not in CORONA_OPTIONS:
            raise ValueError('`extra_descriptions`: {} is not in {}'.format(
                extra_descriptions, CORONA_OPTIONS))

        f = os.path.join(
            TRAINED_QNET_DIR, 
            qnet_type, 
            extra_descriptions + '.joblib')

    elif qnet_type == 'influenza':
        extra_descriptions = extra_descriptions.split(';')
        if len(extra_descriptions) != 3:
            raise ValueError('There must be 3 different descriptions for influenza')

        type_ = extra_descriptions[0]
        if type_ not in INFLUENZA_TYPES:
            raise ValueError('{} is not in {}'.format(type_, INFLUENZA_TYPES))

        protein = extra_descriptions[1]
        if protein not in INFLUENZA_PROTEINS:
            raise ValueError('{} is not in {}'.format(protein, INFLUENZA_PROTEINS))

        base_dir = os.path.join(
            TRAINED_QNET_DIR, 
            qnet_type, 
            '{}human{}/'.format(type_, protein.upper()))

        possible_years = _get_possible_years(base_dir)

        year = extra_descriptions[2]

        if not year.isdigit():
            raise ValueError('Year must be an positive integer.')
        
        year = int(year)

        if year not in possible_years:
            raise ValueError('{} is not in {}'.format(year, possible_years))

        f = os.path.join(base_dir, '{}_{}.joblib'.format(year, year + 1))

    else:
        raise ValueError

    trained_qnet = qnet.load_qnet(f)

    return trained_qnet

def load_sequence(file):
    """Load a fasta sequence from file.

    Parameters
    ----------
    file : str
        File of a fasta sequence

    Returns
    -------
    seq : 1d-like array
        A fasta sequence
    """

    fastas = list(SeqIO.parse(file, "fasta"))

    if len(fastas) != 1:
        raise ValueError('Your fasta sequence can only have 1 sequence in it.')

    seq = np.array(list(str(fastas[0].seq)))
    return seq
                         
                         
    


class Node(object):
    """Decision node in tree

    Parameters
    ----------
    col : int
        Integer indexing the location of feature or column

    col_pval : float
        Probability value from permutation test for feature selection

    lthreshold : list
        List of items for taking the left edge down the tree

    rthreshold : list
        List of items for taking the right edge down the tree

    impurity : float
        Impurity measuring quality of split

    value : 1d array-like or float
        For classification trees, estimate of each class probability
        For regression trees, central tendency estimate

    left : Node
        Another Node

    right : Node
        Another Node

    label_frequency: dict
        Dictionary mapping label to its frequency
    """

    def __init__(self, 
                 col=None, 
                 col_pval=None, 
                 lthreshold=None,
                 rthreshold=None, 
                 impurity=None,
                 value=None, 
                 left=None, 
                 right=None, 
                 label_frequency=None):

        assert isinstance(left, Node) or left is None
        assert isinstance(right, Node) or right is None

        self.col = col
        self.col_pval = col_pval
        self.lthreshold = lthreshold
        self.rthreshold = rthreshold
        self.impurity = impurity
        self.value = value
        self.left = left
        self.right = right
        self.label_frequency = label_frequency

    def __repr__(self):
        return "Node(col={})".format(self.col)

    def __str__(self):
        return self.__repr__()


def get_nodes(root, get_leaves=True, get_non_leaves=True):
    """Traverse a tree and get all the nodes.

    TODO: may need to change this into an iterator for speed purposes

    If `get_leaves` and `get_non_leaves` are both `True`, then
    we will get all the nodes.

    Parameters
    ----------
    root : Node
        root node of the tree

    get_leaves : bool
        If true, we get leaf nodes

    get_non_leaves : bool
        If true, we get non leaf nodes.


    Returns
    -------
    output : list
        list of `Node`
    """

    current = root
    stack = []
    all_nodes = []
    
    if get_leaves and get_non_leaves:
        get_type = 'all'
    elif get_leaves and not get_non_leaves:
        get_type = 'leaf'
    elif not get_leaves and get_non_leaves:
        get_type = 'nonleaf'
    else:
        raise ValueError('We must get either leaves or non-leaves or both.')

    while True: 
        if current is not None: 
            stack.append(current) 
            current = current.left  

        elif(stack): 
            current = stack.pop() 

            if get_type == 'all':
                all_nodes.append(current)
            elif get_type == 'leaf':
                if current.left is None and current.right is None:
                    all_nodes.append(current)
            elif get_type == 'nonleaf':
                if current.left is not None or current.right is not None:
                    all_nodes.append(current)
            else:
                raise ValueError

            current = current.right  

        else: 
            break

    return all_nodes


import multiprocessing
import threading
import warnings

from joblib import delayed, Parallel
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.preprocessing import OrdinalEncoder

warnings.simplefilter('ignore')

# Package imports
from .feature_selectors import (permutation_test_mc, permutation_test_mi,
                               permutation_test_dcor, permutation_test_pcor,
                               permutation_test_rdc, permutation_test_chi2)
from .feature_selectors import mc_fast, mi, pcor, py_dcor
from .scorers import gini_index, mse
from .utils import bayes_boot_probs, logger, powerset
from .citrees import CITreeClassifier



def stratify_sampled_idx(random_state, y, bayes):
    """Indices for stratified bootstrap sampling in classification

    Parameters
    ----------
    random_state : int
        Sets seed for random number generator

    y : 1d array-like
        Array of labels

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    Returns
    -------
    idx : list
        Stratified sampled indices for each class
    """
    np.random.seed(random_state)
    idx = []
    for label in np.unique(y):

        # Grab indices for class
        tmp = np.where(y==label)[0]

        # Bayesian bootstrapping if specified
        p = bayes_boot_probs(n=len(tmp)) if bayes else None

        idx.append(np.random.choice(tmp, size=len(tmp), replace=True, p=p))

    return idx


def stratify_unsampled_idx(random_state, y, bayes):
    """Unsampled indices for stratified bootstrap sampling in classification

    Parameters
    ----------
    random_state : int
        Sets seed for random number generator

    y : 1d array-like
        Array of labels

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    Returns
    -------
    idx : list
        Stratified unsampled indices for each class
    """
    np.random.seed(random_state)
    sampled = stratify_sampled_idx(random_state, y, bayes)
    idx     = []
    for i, label in enumerate(np.unique(y)):
        idx.append(np.setdiff1d(np.where(y==label)[0], sampled[i]))
    return idx


def balanced_sampled_idx(random_state, y, bayes, min_class_p):
    """Indices for balanced bootstrap sampling in classification

    Parameters
    ----------
    random_state : int
        Sets seed for random number generator

    y : 1d array-like
        Array of labels

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    min_class_p : float
        Minimum proportion of class labels

    Returns
    -------
    idx : list
        Balanced sampled indices for each class
    """
    np.random.seed(random_state)
    idx, n = [], int(np.floor(min_class_p*len(y)))
    for i, label in enumerate(np.unique(y)):

        # Grab indices for class
        tmp = np.where(y==label)[0]

        # Bayesian bootstrapping if specified
        p = bayes_boot_probs(n=len(tmp)) if bayes else None

        idx.append(np.random.choice(tmp, size=n, replace=True, p=p))

    return idx


def balanced_unsampled_idx(random_state, y, bayes, min_class_p):
    """Unsampled indices for balanced bootstrap sampling in classification

    Parameters
    ----------
    random_state : int
        Sets seed for random number generator

    y : 1d array-like
        Array of labels

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    min_class_p : float
        Minimum proportion of class labels

    Returns
    -------
    idx : list
        Balanced unsampled indices for each class
    """
    np.random.seed(random_state)
    sampled = balanced_sampled_idx(random_state, y, bayes, min_class_p)
    idx     = []
    for i, label in enumerate(np.unique(y)):
        idx.append(np.setdiff1d(np.where(y==label)[0], sampled[i]))
    return idx


def normal_sampled_idx(random_state, n, bayes):
    """Indices for bootstrap sampling

    Parameters
    ----------
    random_state : int
        Sets seed for random number generator

    n : int
        Sample size

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    Returns
    -------
    idx : list
        Sampled indices
    """
    np.random.seed(random_state)

    # Bayesian bootstrapping if specified
    p = bayes_boot_probs(n=n) if bayes else None

    return np.random.choice(np.arange(n, dtype=int), size=n, replace=True, p=p)


def normal_unsampled_idx(random_state, n, bayes):
    """Unsampled indices for bootstrap sampling

    Parameters
    ----------
    random_state : int
        Sets seed for random number generator

    y : 1d array-like
        Array of labels

    n : int
        Sample size

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    Returns
    -------
    idx : list
        Unsampled indices
    """
    sampled = normal_sampled_idx(random_state, n, bayes)
    counts  = np.bincount(sampled, minlength=n)
    return np.arange(n, dtype=int)[counts==0]


def _parallel_fit_classifier(tree, X, y, n, tree_idx, n_estimators, bootstrap,
                             bayes, verbose, random_state, class_weight=None,
                             min_dist_p=None):
    """Utility function for building trees in parallel

    Note: This function can't go locally in a class, because joblib complains
          that it cannot pickle it when placed there

    Parameters
    ----------
    tree : CITreeClassifier
        Instantiated conditional inference tree

    X : 2d array-like
        Array of features

    y : 1d array-like
        Array of labels

    n : int
        Number of samples

    tree_idx : int
        Index of tree in forest

    n_estimators : int
        Number of total estimators

    bootstrap : bool
        Whether to perform bootstrap sampling

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    verbose : bool or int
        Controls verbosity of training process

    random_state : int
        Sets seed for random number generator

    class_weight : str
        Type of sampling during bootstrap, None for regular bootstrapping,
        'balanced' for balanced bootstrap sampling, and 'stratify' for
        stratified bootstrap sampling

    min_class_p : float
        Minimum proportion of class labels

    Returns
    -------
    tree : CITreeClassifier
        Fitted conditional inference tree
    """
    # Print status if conditions met
    if verbose and n_estimators >= 10:
        denom = n_estimators if verbose > 1 else 10
        if (tree_idx+1) % int(n_estimators/denom) == 0:
            logger("tree", "Building tree %d/%d" % (tree_idx+1, n_estimators))

    # Bootstrap sample if specified
    if bootstrap:
        random_state = random_state*(tree_idx+1)
        if class_weight == 'balanced':
            idx = np.concatenate(
                balanced_sampled_idx(random_state, y, bayes, min_dist_p)
                )
        elif class_weight == 'stratify':
            idx = np.concatenate(
                stratify_sampled_idx(random_state, y, bayes)
                )
        else:
            idx = normal_sampled_idx(random_state, n, bayes)

        # Note: We need to pass the classes in the case of the bootstrap
        # because not all classes may be sampled and when it comes to prediction,
        # the tree models learns a different number of classes across different
        # bootstrap samples
        tree.fit(X[idx], y[idx], np.unique(y))
    else:
        tree.fit(X, y)
    
    return tree



def _accumulate_prediction(predict, X, out, lock):
    """Utility function to aggregate predictions in parallel

    Parameters
    ----------
    predict : function handle
        Alias to prediction method of class

    X : 2d array-like
        Array of features

    out : 1d or 2d array-like
        Array of labels

    lock : threading lock
        A lock that controls worker access to data structures for aggregating
        predictions

    Returns
    -------
    None
    """
    prediction = predict(X)
    with lock:
        if len(out) == 1:
            out[0] += prediction
        else:
            # TODO: need to check this because we used the range from external
            # need to check all otther ranges as well
            for i in range(len(out)): out[i] += prediction[i]


class CIForestClassifier(BaseEstimator, ClassifierMixin):
    """Conditional forest classifier

    Parameters
    ----------
    min_samples_split : int
        Minimum samples required for a split

    alpha : float
        Threshold value for selecting feature with permutation tests. Smaller
        values correspond to shallower trees

    selector : str
        Variable selector for finding strongest association between a feature
        and the label

    max_depth : int
        Maximum depth to grow tree

    max_feats : str or int
        Maximum feats to select at each split. String arguments include 'sqrt',
        'log', and 'all'

    n_permutations : int
        Number of permutations during feature selection

    early_stopping : bool
        Whether to implement early stopping during feature selection. If True,
        then as soon as the first permutation test returns a p-value less than
        alpha, this feature will be chosen as the splitting variable

    muting : bool
        Whether to perform variable muting

    verbose : bool or int
        Controls verbosity of training and testing

    bootstrap : bool
        Whether to perform bootstrap sampling for each tree

    bayes : bool
        If True, performs Bayesian bootstrap sampling

    class_weight : str
        Type of sampling during bootstrap, None for regular bootstrapping,
        'balanced' for balanced bootstrap sampling, and 'stratify' for
        stratified bootstrap sampling

    n_jobs : int
        Number of jobs for permutation testing

    random_state : int
        Sets seed for random number generator
    """
    def __init__(self, min_samples_split=2, alpha=.05, selector='mc', max_depth=-1,
                 n_estimators=100, max_feats='sqrt', n_permutations=100,
                 early_stopping=True, muting=True, verbose=0, bootstrap=True,
                 bayes=True, class_weight='balanced', n_jobs=-1, random_state=None):

        # Error checking
        if alpha <= 0 or alpha > 1:
            raise ValueError("Alpha (%.2f) should be in (0, 1]" % alpha)
        if selector not in ['mc', 'mi', 'hybrid']:
            raise ValueError("%s not a valid selector, valid selectors are " \
                             "mc, mi, and hybrid")
        if n_permutations < 0:
            raise ValueError("n_permutations (%s) should be > 0" % \
                             str(n_permutations))
        if not isinstance(max_feats, int) and max_feats not in ['sqrt', 'log', 'all', -1]:
            raise ValueError("%s not a valid argument for max_feats" % \
                             str(max_feats))
        if n_estimators < 0:
            raise ValueError("n_estimators (%s) must be > 0" % \
                             str(n_estimators))

        # Only for classifier model
        if class_weight not in [None, 'balanced', 'stratify']:
            raise ValueError("%s not a valid argument for class_weight" % \
                             str(class_weight))

        # Placeholder variable for regression model (not applicable)
        if class_weight is None: self.min_class_p = None

        # Define attributes
        self.alpha             = float(alpha)
        self.selector          = selector
        self.min_samples_split = max(1, min_samples_split)
        self.n_permutations    = int(n_permutations)
        if max_depth == -1:
            self.max_depth = max_depth
        else:
            self.max_depth = int(max(1, max_depth))

        self.n_estimators   = int(max(1, n_estimators))
        self.max_feats      = max_feats
        self.bootstrap      = bootstrap
        self.early_stopping = early_stopping
        self.muting         = muting
        self.n_jobs         = n_jobs
        self.verbose        = verbose
        self.class_weight   = class_weight
        self.bayes          = bayes

        if random_state is None:
            self.random_state = np.random.randint(1, 9999)
        else:
            # TODO: ADD CHECK FOR CRAZY LARGE INTEGER?
            self.random_state = int(random_state)

        # Package params for calling CITreeClassifier
        self.params = {
            'alpha'             : self.alpha,
            'selector'          : self.selector,
            'min_samples_split' : self.min_samples_split,
            'n_permutations'    : self.n_permutations,
            'max_feats'         : self.max_feats,
            'early_stopping'    : self.early_stopping,
            'muting'            : self.muting,
            'verbose'           : 0,
            'n_jobs'            : 1,
            'random_state'      : None,
            }


    def fit(self, X, y):
        """Fit conditional forest classifier

        Parameters
        ----------
        X : 2d array-like
            Array of features

        y : 1d array-like
            Array of labels

        Returns
        -------
        self : CIForestClassifier
            Instance of CIForestClassifier
        """
        self.labels_    = np.unique(y)
        self.n_classes_ = len(self.labels_)

        if self.verbose:
            logger("tree", "Training ensemble with %d trees on %d samples" % \
                    (self.n_estimators, X.shape[0]))

        # Instantiate base tree models
        self.estimators_ = []
        for i in range(self.n_estimators):
            self.params['random_state'] = self.random_state*(i+1)
            self.estimators_.append(CITreeClassifier(**self.params))

        # Define class distribution
        self.class_dist_p = np.array([
                np.mean(y==label) for label in np.unique(y)
            ])

        # Train models
        n = X.shape[0]
        self.estimators_ = \
            Parallel(n_jobs=self.n_jobs, backend='loky')(
            delayed(_parallel_fit_classifier)(
                self.estimators_[i], X, y, n, i, self.n_estimators,
                self.bootstrap, self.bayes, self.verbose, self.random_state,
                self.class_weight, np.min(self.class_dist_p)
                )
            for i in range(self.n_estimators)
            )

        # Accumulate feature importances (mean decrease impurity)
        self.feature_importances_ = np.sum([
                tree.feature_importances_ for tree in self.estimators_],
                axis=0
            )
        sum_fi = np.sum(self.feature_importances_)
        if sum_fi > 0: self.feature_importances_ /= sum_fi

        return self


    def predict_proba(self, X):
        """Predicts class probabilities for feature vectors X

        Parameters
        ----------
        X : 2d array-like
            Array of features

        Returns
        -------
        class_probs : 2d array-like
            Array of predicted class probabilities
        """
        if self.verbose:
            logger("test", "Predicting labels for %d samples" % X.shape[0])

        # Parallel prediction
        all_proba = np.zeros((X.shape[0], self.n_classes_), dtype=np.float64)
        lock      = threading.Lock()
        Parallel(n_jobs=self.n_jobs, backend="threading")(
            delayed(_accumulate_prediction)(e.predict_proba, X, all_proba, lock)
            for e in self.estimators_)

        # Normalize probabilities
        all_proba /= len(self.estimators_)
        if len(all_proba) == 1:
            return all_proba[0]
        else:
            return all_proba


    def predict(self, X):
        """Predicts class labels for feature vectors X

        Parameters
        ----------
        X : 2d array-like
            Array of features

        Returns
        -------
        y : 1d array-like
            Array of predicted classes
        """
        y_proba = self.predict_proba(X)
        return np.argmax(y_proba, axis=1)


from decimal import Decimal

# from numba import autojit
import numpy as np

# from externals.six.moves import range


def powerset(s):
    """Get the power set of a list or set.
    """
    
    x = len(s)
    masks = [1 << i for i in range(x)]
    for i in range(1 << x):
        yield [ss for mask, ss in zip(masks, s) if i & mask]


def remove_zeros(r, axis):
    """Remove rows along a certain axis where the value is all zero.
    """

    if axis == 0:
        return r[~np.all(r == 0, axis=1)]
    elif axis == 1:
        return r[:, ~np.all(r == 0, axis=0)]
    else:
        raise ValueError('not a correct axis that we can use.')

def scientific_notation(num):
    """Convert a number into scientific notation

    Parameters
    ----------
    num : float
        Any number
    
    Returns
    -------
    output : str
        String representation of the number
    """

    return "{:.2E}".format(Decimal(num))

def bayes_boot_probs(n):
    """Bayesian bootstrap sampling for case weights
    
    Parameters
    ----------
    n : int
        Number of Bayesian bootstrap samples
    
    Returns
    -------
    p : 1d array-like
        Array of sampling probabilities
    """
    p = np.random.exponential(scale=1.0, size=n)
    return p/p.sum()


# @autojit(nopython=True, cache=True, nogil=True)
def auc_score(y_true, y_prob):
    """ADD
    
    Parameters
    ----------
    
    Returns
    -------
    """
    y_true, n   = y_true[np.argsort(y_prob)], len(y_true)
    nfalse, auc = 0, 0.0
    for i in range(n):
        nfalse += 1 - y_true[i]
        auc    += y_true[i] * nfalse
    auc /= (nfalse * (n - nfalse))
    return auc


def logger(name, message):
    """Prints messages with style "[NAME] message"
    
    Parameters
    ----------
    name : str
        Short title of message, for example, train or test

    message : str
        Main description to be displayed in terminal
    
    Returns
    -------
    None
    """
    print('[{name}] {message}'.format(name=name.upper(), message=message))


def estimate_margin(y_probs, y_true):
    """Estimates margin function of forest ensemble

    Note : This function is similar to margin in R's randomForest package
    
    Parameters
    ----------
    y_probs : 2d array-like
        Predicted probabilities where each row represents predicted
        class distribution for sample and each column corresponds to 
        estimated class probability

    y_true : 1d array-like
        Array of true class labels
    
    Returns
    -------
    margin : float
        Estimated margin of forest ensemble
    """
    # Calculate probability of correct class
    n, p        = y_probs.shape
    true_probs  = y_probs[np.arange(n, dtype=int), y_true]

    # Calculate maximum probability for incorrect class
    other_probs = np.zeros(n)
    for i in range(n):
        mask            = np.zeros(p, dtype=bool)
        mask[y_true[i]] = True
        other_idx       = np.ma.array(y_probs[i,:], mask=mask).argmax()
        other_probs[i]  = y_probs[i, other_idx]
    
    # Margin is P(y == j) - max(P(y != j))
    return true_probs - other_probs



def assert_array_rank(X, rank):
    """Check if the input is an numpy array and has a certain rank.

    Parameters
    ----------
    X : array-like
        Array to check

    rank : int
        Rank of the tensor to check
    
    Returns
    -------
    None
    """

    if not isinstance(X, np.ndarray):
        raise ValueError('You must pass in a numpy array!')

    if len(X.shape) != rank:
        raise ValueError('You must pass in a {}-rank array!'.format(rank))

def assert_string_type(X, name):
    """Check if the input is of string datatype.

    Parameters
    ----------
    X : array-like
        Array to check

    name : str
        Name of the input
    
    Returns
    -------
    None
    """

    if not np.issubdtype(X.dtype, np.str_):
        raise ValueError('{} must contain only strings!'.format(name))

def sample_from_dict(distrib):
    """Choose an item from the distribution

    Parameters
    ----------
    distrib : dict
        Dictionary mapping keys to its probability values

    Returns
    -------
    item : key of dict
        A chosen key from the dictionary
    """

    keys = []
    probs = []
    for k, prob in distrib.items():
        keys.append(k)
        probs.append(prob)

    probs = np.array(probs)
    probs /= probs.sum()

    item = np.random.choice(keys, p=probs)

    return item



import warnings
from collections import Counter, defaultdict

from joblib import delayed, Parallel
import numpy as np
from sklearn.base import BaseEstimator, ClassifierMixin
from sklearn.preprocessing import OrdinalEncoder

warnings.simplefilter('ignore')

from .feature_selectors import permutation_test_chi2
from .scorers import gini_index
from .utils import logger, powerset
from .tree import Node, get_nodes
from ._config import get_config
from ._encoders import OrdinalEncoderWithNaN

class CITreeBase(object):
    """Base class for conditional inference tree. 

    Parameters
    ----------
    min_samples_split : int
        Minimum samples required for a split

    alpha : float
        Threshold value for selecting feature with permutation tests. Smaller
        values correspond to shallower trees

    max_depth : int
        Maximum depth to grow tree

    max_feats : str or int
        Maximum feats to select at each split. String arguments include 'sqrt',
        'log', and 'all'

    n_permutations : int
        Number of permutations during feature selection

    early_stopping : bool
        Whether to implement early stopping during feature selection. If True,
        then as soon as the first permutation test returns a p-value less than
        alpha, this feature will be chosen as the splitting variable

    muting : bool
        Whether to perform variable muting

    verbose : bool or int
        Controls verbosity of training and testing

    random_state : int
        Sets seed for random number generator
    """

    def __init__(
        self, 
        min_samples_split=2, 
        alpha=.05, 
        max_depth=-1,
        max_feats=-1, 
        n_permutations=100, 
        early_stopping=False,
        muting=True, 
        verbose=0, 
        random_state=None):

        if alpha <= 0 or alpha > 1:
            raise ValueError("Alpha (%.2f) should be in (0, 1]" % alpha)

        if (not isinstance(n_permutations, int)) or (n_permutations < 0):
            raise ValueError('The number of permutations must be a postive integer.')

        if not isinstance(max_feats, int) and max_feats not in ['sqrt', 'log', 'all', -1]:
            raise ValueError("%s not a valid argument for max_feats" % \
                             str(max_feats))


        self.alpha = float(alpha)
        self.min_samples_split = max(1, int(min_samples_split))
        self.n_permutations = int(n_permutations)
        self.max_feats = max_feats
        self.early_stopping = early_stopping
        self.muting = muting
        self.verbose = verbose
        self.root = None
        self.splitter_counter_ = 0

        if max_depth == -1:
            self.max_depth = np.inf
        elif max_depth == 0:
            raise ValueError('The depth of the tree can not be zero.')
        elif not isinstance(max_depth, int):
            raise ValueError('You must provide an integer as the depth.')
        elif max_depth < -1:
            raise ValueError('The maximum depth cannot be less than -1.')
        else:
            self.max_depth = max_depth

        if random_state is None:
            self.random_state = np.random.randint(1, 9999)
        else:
            self.random_state = int(random_state)
            if self.random_state > 9999:
                raise ValueError('The random state must be less than 10000.')


    def _mute_feature(self, col_to_mute):
        """Removes variable from being selected

        Parameters
        ----------
        col_to_mute : int
            Integer index of column to remove
        """
        # Remove feature from protected features array
        idx = np.where(self.available_features_ == col_to_mute)[0]

        # Mute feature if not in protected set
        if idx in self.protected_features_:
            return
        else:
            self.available_features_ = np.delete(self.available_features_, idx)

            # Recalculate actual number for max_feats before fitting
            p = self.available_features_.shape[0]
            if self.max_feats == 'sqrt':
                self.max_feats = int(np.sqrt(p))
            elif self.max_feats == 'log':
                self.max_feats = int(np.log(p+1))
            elif self.max_feats in ['all', -1]:
                self.max_feats = p
            else:
                self.max_feats = int(self.max_feats)

            # Check to make sure max_feats is not larger than the number of remaining
            # features
            if self.max_feats > len(self.available_features_):
                self.max_feats = len(self.available_features_)


    def _selector(self, X, y, col_idx):
        """Find feature most correlated with label"""
        raise NotImplementedError("_splitter method not callable from base class")


    def _splitter(self, *args, **kwargs):
        """Finds best split for feature"""
        raise NotImplementedError("_splitter method not callable from base class")


    def _build_tree(self, X, y, depth=0):
        """Recursively builds tree

        Parameters
        ----------
        X : 2d array-like
            Array of features

        y : 1d array-like
            Array of labels

        depth : int
            Depth of current recursive call

        Returns
        -------
        Node : object
            Child node or terminal node in recursive splitting
        """
        n, p = X.shape

        # Check for stopping criteria
        if n > self.min_samples_split and \
            depth < self.max_depth and \
            not np.all(y == y[0]):

            # Controls randomness of column sampling
            self.splitter_counter_ += 1
            np.random.seed(self.random_state*self.splitter_counter_)

            # Find column with strongest association with outcome
            try:
                col_idx = np.random.choice(
                    self.available_features_,
                    size=self.max_feats, 
                    replace=False)
            except:
                col_idx = np.random.choice(
                    self.available_features_,
                    size=len(self.available_features_),
                    replace=False)

            col, col_pval = self._selector(X, y, col_idx)
            
            # Add selected feature to protected features
            if col not in self.protected_features_:
                self.protected_features_.append(col)
                if self.verbose > 1:
                    logger("tree", "Added feature %d to protected set, size "
                           "= %d" % (col, len(self.protected_features_)))

            if col_pval <= self.alpha:

                # Find best split among selected variable
                impurity, lthreshold, rthreshold, left, right = self._splitter(X, y, n, col)
                if left and right and len(left[0]) > 0 and len(right[0]) > 0:

                    # Build subtrees for the right and left branches
                    if self.verbose:
                        logger("tree", "Building left subtree with "
                                       "%d samples at depth %d" % \
                                       (len(left[0]), depth+1))
                    left = self._build_tree(*left, depth=depth+1)

                    if self.verbose:
                        logger("tree", "Building right subtree with "
                                       "%d samples at depth %d" % \
                                        (len(right[0]), depth+1))
                    right = self._build_tree(*right, depth=depth+1)

                    # Return all arguments to constructor except value
                    node = Node(
                        col=col, 
                        col_pval=col_pval, 
                        lthreshold=lthreshold,
                        rthreshold=rthreshold,
                        left=left, 
                        right=right,
                        impurity=impurity, 
                        label_frequency=Counter(y))
                    return node

        # Calculate terminal node value
        if self.verbose: 
            logger("tree", "Root node reached at depth %d" % depth)

        value = self.node_estimate(y)

        # return terminal node because can no longer find a split
        return Node(value=value, label_frequency=Counter(y))


    def fit(self, X, y=None):
        """Train model. 
        
        X, y must contain only string datatypes.

        Parameters
        ----------
        X : 2d array-like
            Array of categorical features

        y : 1d array-like
            Array of labels

        Returns
        -------
        self : CITreeBase
            Instance of CITreeBase class
        """

        if not isinstance(X, np.ndarray) or len(X.shape) != 2:
            raise ValueError('X must be a 2d numpy array!')

        if not np.issubdtype(X.dtype, np.str_):
            raise ValueError('X must contain only strings! This is because this implementation only allows for categorical inputs.')

        if not isinstance(y, np.ndarray) or len(y.shape) != 1:
            raise ValueError('y must be a 1d numpy array!')

        if self.verbose:
            logger("tree", "Building root node with %d samples" % X.shape[0])

        # Calculate actual number for max_feats before fitting
        p = X.shape[1]
        if self.max_feats == 'sqrt':
            self.max_feats = int(np.sqrt(p))
        elif self.max_feats == 'log':
            self.max_feats = int(np.log(p+1))
        elif self.max_feats in ['all', -1]:
            self.max_feats = p
        else:
            self.max_feats = int(self.max_feats)

        # Begin recursive build
        self.protected_features_ = []
        self.available_features_ = np.arange(p, dtype=int)
        self.feature_importances_ = np.zeros(p)
        self.root = self._build_tree(X, y)
        sum_fi = np.sum(self.feature_importances_)
        if sum_fi > 0: 
            self.feature_importances_ /= sum_fi

        return self


    def predict_label(self, X, tree=None):
        """Predicts label

        Parameters
        ----------
        X : 1d array-like
            Array of features for single sample

        tree : CITreeBase
            Trained tree

        Returns
        -------
        label : str
            Predicted label
        """
        # If we have a value => return value as the prediction
        if tree is None: 
            tree = self.root
        if tree.value is not None: 
            return tree.value

        # Determine if we will follow left or right branch
        feature_value = X[tree.col]

        lthreshold = tree.lthreshold
        rthreshold = list(tree.rthreshold) + [get_config()['nan_value']]
        if feature_value in lthreshold:
            branch = tree.left
        elif feature_value in rthreshold:
            branch = tree.right
        else:
            raise ValueError('{} is not in: {} or in: {}'.format(feature_value,
                                                                 lthreshold,
                                                                 rthreshold))

        return self.predict_label(X, branch)


    def predict(self, *args, **kwargs):
        """Predicts labels on test data. This method should not be 
        callable from base class.
        """
        
        raise NotImplementedError("predict method not callable from base class")


    def print_tree(self, tree=None, indent=" ", child=None):
        """Prints tree structure

        Parameters
        ----------
        tree : CITreeBase
            Trained tree model

        indent : str
            Indent spacing

        child : Node
            Left or right child node

        Returns
        -------
        None
        """
        # If we're at leaf => print the label
        if not tree: tree = self.root
        if tree.value is not None: print("label:", tree.value)

        # Go deeper down the tree
        else:
            # Print splitting rule
            print("X[:,%s] %s %s " % (tree.col,
                                      '<=' if child in [None, 'left'] else '>',
                                      tree.lthreshold))

            # Print the left child
            print("%sL: " % (indent), end="")
            self.print_tree(tree.left, indent + indent, 'left')

            # Print the right
            print("%sR: " % (indent), end="")
            self.print_tree(tree.right, indent + indent, 'right')


class CITreeClassifier(CITreeBase, BaseEstimator, ClassifierMixin):
    """Conditional inference tree classifier

    NOTE: as of now, the features can only be categorical

    Parameters
    ----------
    selector : str
        Variable selector for finding strongest association between a feature
        and the label

    Derived from CITreeBase class; see constructor for parameter definitions
    """

    def __init__(
        self,
        min_samples_split=2,
        alpha=.05,
        selector='chi2',
        max_depth=-1,
        max_feats=-1,
        n_permutations=100,
        early_stopping=False,
        muting=True,
        verbose=0,
        random_state=None):

        # Define node estimate
        self.node_estimate = self._estimate_proba

        self.selector = selector

        if selector == 'chi2':
            self._selector = self._cor_selector
            self._perm_test = permutation_test_chi2
        else:
            raise ValueError('Not a correct selector: {}'.format(selector))

        super(CITreeClassifier, self).__init__(min_samples_split=min_samples_split,
                                               alpha=alpha,
                                               max_depth=max_depth,
                                               max_feats=max_feats,
                                               n_permutations=n_permutations,
                                               early_stopping=early_stopping,
                                               muting=muting,
                                               verbose=verbose,
                                               random_state=random_state)


    def _split_by_gini(self, X, y):
        """Given a single covariate, split the data to minimize the 
        gini coefficient.

        TODO: do randomly sampling if the number of categories is large

        Parameters
        ----------
        X: 2d array-like of shape (n, 1)

        y: 1d array-like
        """

        # skip splitting because there's no way we can split the data
        # so both subsets have size less than `min_samples_split` 
        if y.shape[0] < 2 * self.min_samples_split:
            return None, None

        index_to_gini_values = {}
        index_to_subset = {}

        unique_X = np.sort(np.unique(X))
        num_unique = len(unique_X)
        
        # if the number of unique possibilities is too large, we need to randomly
        # sample because the powerset grows exponentially
        if num_unique <= 6:
            subsets = list(powerset(unique_X))
        else:
            # TODO: need to replace this with a generator
            subsets = []
            # NOTE: there may be some inefficiency because when randomly sampling,
            # we may introduce redundant splits
            for i in range(2 ** 6):
                randint = np.random.randint(low=1, high=num_unique)
                indices = np.random.choice(
                    num_unique, 
                    size=randint,
                    replace=False)
                    
                subset = unique_X[indices]
                subsets.append(subset)

        for i, subset in enumerate(subsets):
            # TODO: sort subset

            # ignore the case where the list is empty or the subset
            # is the set itself because no split is actually made
            if len(subset) == 0:
                continue 
            elif len(subset) == len(unique_X):
                continue

            subset_complement = unique_X[~np.isin(unique_X, subset)]
            y1 = y[np.isin(X, subset)]
            y2 = y[np.isin(X, subset_complement)]

            # ignore splits that create small subsets
            if y1.shape[0] < self.min_samples_split or y2.shape[0] < self.min_samples_split:
                continue

            gini_value = gini_index(y1, subset) + gini_index(y2, subset_complement)

            index_to_gini_values[i] = gini_value
            index_to_subset[i] = subset

        if len(index_to_subset) == 0:
            return None, None

        min_key = min(index_to_gini_values, key=index_to_gini_values.get)

        lthreshold = index_to_subset[min_key]
        rthreshold = unique_X[~np.isin(unique_X, lthreshold)]
        return lthreshold, rthreshold 


    def _splitter(self, X, y, n, col):
        """Splits data set into two child nodes based on optimized weighted
        gini index

        Parameters
        ----------
        X : 2d array-like
            Array of features

        y : 1d array-like
            Array of labels

        n : int
            Number of samples

        col : list
            Column of X to search for best split

        Returns
        -------
        best_impurity : float
            Gini index associated with best split

        best_threshold : float
            X value associated with splitting of data set into two child nodes

        left : tuple
            Left child node data consisting of two elements: (features, labels)

        right : tuple
            Right child node data consisting of two elements: (features labels)
        """

        if self.verbose > 1:
            logger("splitter", "Testing splits on feature %d" % col)

        # Initialize variables for splitting
        impurity = 0.0
        left, right = None, None

        X_col = X[:, col]
        nan_value = get_config()['nan_value']
        not_nan_indices = X_col != nan_value

        lthreshold, rthreshold = self._split_by_gini(
            X_col[not_nan_indices], 
            y[not_nan_indices])

        # Skip if we cannot find a split
        if lthreshold is None and rthreshold is None:
            return impurity, lthreshold, rthreshold, left, right

        idx = np.where(np.isin(X[:, col], lthreshold), 1, 0)
        
        X_left, y_left   = X[idx==1], y[idx==1]
        X_right, y_right = X[idx==0], y[idx==0]
        n_left, n_right  = X_left.shape[0], X_right.shape[0]

        # Skip small splits
        if n_left < self.min_samples_split or n_right < self.min_samples_split:
            return impurity, lthreshold, rthreshold, left, right

        node_impurity  = gini_index(y, self.labels_)
        left_impurity  = gini_index(y_left, self.labels_)*(n_left/float(n))
        right_impurity = gini_index(y_right, self.labels_)*(n_right/float(n))

        # Define groups and calculate impurity decrease
        left, right = (X_left, y_left), (X_right, y_right)
        impurity    = node_impurity - (left_impurity + right_impurity)

        # Update feature importance (mean decrease impurity)
        self.feature_importances_[col] += impurity

        return impurity, lthreshold, rthreshold, left, right


    def _cor_selector(self, X, y, col_idx):
        """Selects feature most correlated with y using permutation tests with
        a correlation measure

        Parameters
        ----------
        X : 2d array-like
            Array of features

        y : 1d array-like
            Array of labels

        col_idx : list
            Columns of X to examine for feature selection

        Returns
        -------
        best_col : int
            Best column from feature selection. Note, if early_stopping is
            enabled then this may not be the absolute best column

        best_pval : float
            Probability value from permutation test
        """
        # Select random column from start and update
        best_col, best_pval = np.random.choice(col_idx), np.inf

        y = self.y_enc.transform(y.reshape(-1, 1)).reshape(-1)

        # Iterate over columns
        for col in col_idx:
            X_col = X[:, col]
            # Mute feature and continue since constant
            if np.all(X_col == X[0, col]) and len(self.available_features_) > 1:
                self._mute_feature(col)
                if self.verbose: 
                    logger("tree", "Constant values, muting feature %d" % col)
                continue

            X_col = self.X_encs[col].transform(X_col.reshape(-1, 1)).reshape(-1)

            # remove datapoints where X_col is NaN (-1)
            not_nan_cols = X_col != -1

            pval = self._perm_test(x=X_col[not_nan_cols],
                                   y=y[not_nan_cols],
                                   n_classes=self.n_classes_,
                                   B=self.n_permutations,
                                   random_state=self.random_state)

            # If variable muting
            if self.muting and \
               pval == 1.0 and \
               self.available_features_.shape[0] > 1:
                self._mute_feature(col)
                if self.verbose: logger("tree", "ASL = 1.0, muting feature %d" % col)

            if pval < best_pval:
                best_col, best_pval = col, pval

                # If early stopping
                if self.early_stopping and best_pval < self.alpha:
                    if self.verbose: logger("tree", "Early stopping")
                    return best_col, best_pval

        return best_col, best_pval


    def _estimate_proba(self, y):
        """Estimates class distribution in node

        Parameters
        ----------
        y : 1d array-like
            Array of labels

        Returns
        -------
        class_probs : 1d array-like
            Array of class probabilities
        """
        return np.array([np.mean(y == label) for label in self.labels_])


    def fit(self, X, y, labels=None):
        """Train conditional inference tree classifier

        Parameters
        ----------
        X : 2d array-like
            Array of features

        y : 1d array-like
            Array of labels

        labels : 1d array-like
            Array of unique class labels

        Returns
        -------
        self : CITreeClassifier
            Instance of CITreeClassifier class
        """

        if labels is None:
            labels = 'auto'

        nan_value = get_config()['nan_value']

        # remove datapoints where y is NaN
        label_indices = y != nan_value
        y = y[label_indices]
        X = X[label_indices]

        if y.shape[0] != X.shape[0]:
            raise ValueError('`X` and `y` do not have the same number of samples.')

        self.y_enc = OrdinalEncoder(categories=labels, dtype=np.int32)

        self.y_enc.fit(y.reshape(-1, 1))

        self.X_encs = {}
        for col_index in np.arange(X.shape[1]):
            enc = OrdinalEncoderWithNaN(dtype=np.int32)
            enc.fit(X[:, col_index].reshape(-1, 1))
            self.X_encs[col_index] = enc

        self.labels_ = self.y_enc.categories_[0]
        self.n_classes_ = len(self.labels_)
        super(CITreeClassifier, self).fit(X, y)
        return self


    def predict_proba(self, X):
        """Predicts class probabilities for feature vectors X

        Parameters
        ----------
        X : 2d array-like
            Array of features

        Returns
        -------
        class_probs : 2d array-like
            Array of predicted class probabilities
        """
        if self.verbose:
            logger("test", "Predicting labels for %d samples" % X.shape[0])

        num_samples = X.shape[0]
        predictions = np.zeros((num_samples, self.n_classes_))
        for i in np.arange(0, num_samples):
            predictions[i] = self.predict_label(X[i])

        return predictions


    def predict(self, X):
        """Predicts class labels for feature vectors X

        Parameters
        ----------
        X : 2d array-like
            Array of features

        Returns
        -------
        y : 1d array-like
            Array of predicted classes
        """
        
        # TODO: treat a feature as missing data if that feature is
        # not an instance of the corresponding training feature

        y_proba = self.predict_proba(X)
        ordinal_prediction = np.argmax(y_proba, axis=1).reshape((-1, 1))
        categorical_prediction = self.y_enc.inverse_transform(ordinal_prediction)
        y = categorical_prediction.reshape(-1)
        return y
        


def get_feature_importance(citree, normalize=True):
    """Get the feature importance of the citree.

    Parameters
    ----------
    citree : CITreeBase
        A conditional inference tree

    normalize : bool
        Whether to normalize the feature importance or not

    Returns
    -------
    col_to_importance : dict
        Mapping from column index to total feature importance
    """

    # if not isinstance(citree, CITreeClassifier):
    #     raise ValueError('The input must be a `CITreeClassifier`.')

    nodes = get_nodes(citree.root, get_leaves=False)

    col_to_importance_ = defaultdict(lambda: 0)

    for node in nodes:
        col_to_importance_[node.col] += node.impurity

    if normalize:
        total = sum(col_to_importance_.values())
        col_to_importance = {k: float(v) / total for k, v in col_to_importance_.items()}
    else:
        col_to_importance = dict(col_to_importance_)

    return col_to_importance







"""Global configuration state and functions for management

TODO: allow for config values to be set by users
"""

_global_config = {
    'nan_value': '',
}


def get_config():
    """Retrieve current values for configuration set by :func:`set_config`

    Parameters
    ----------
    None

    Returns
    -------
    config : dict
        Dictionary mapping config key to its value
    """

    return _global_config.copy()


import unittest

import pandas as pd

from quasinet.tree import Node, get_nodes

class TestTree(unittest.TestCase):

    @staticmethod
    def initialize_node():
        root = Node(1) 
        root.left = Node(2) 
        root.right = Node(3) 
        root.left.left = Node(4) 
        root.left.right = Node(5) 

        return root

    def test_get_nodes(self):
        root = self.initialize_node()
        nodes = get_nodes(root)
        all_cols = set([node.col for node in nodes])
        self.assertEqual(
            all_cols,
            set(range(1, 6)))

if __name__ == "__main__":
    unittest.main()


cimport numpy as np
import numpy as np

ctypedef np.float32_t float32
ctypedef np.int32_t int32
ctypedef np.int8_t int8

def _combine_two_distribs(dict seq1_distrib, dict seq2_distrib):
    """Combine two distributions together.

    The two distributions may contain different random variables.
    If one random variable doesn't exist in one of the distributions,
    then we set the probability to 0.

    Parameters
    ----------
    seq1_distrib : dict
        Dictionary mapping random variables to probabilities.

    seq2_distrib : dict
        Dictionary mapping random variables to probabilities.

    Returns
    -------
    output : 2d array-like
        array of shape (2, total unique random variables)
    """

    cdef float32 float32_zero = 0.0
    cdef np.ndarray[float32, ndim=2] distrib
    cdef str seq1_response, seq2_response, x
    cdef int32 i

    # make sure both distribution contains the same responses
    for seq1_response in seq1_distrib.keys():
        if seq1_response not in seq2_distrib:
            seq2_distrib[seq1_response] = float32_zero

    for seq2_response in seq2_distrib.keys():
        if seq2_response not in seq1_distrib:
            seq1_distrib[seq2_response] = float32_zero

    distrib = np.empty((2, len(seq2_distrib)), dtype=np.float32)
    for i, x in enumerate(seq1_distrib.keys()):
        distrib[0, i] = seq1_distrib[x]
        distrib[1, i] = seq2_distrib[x]

    return distrib


import random
import warnings

import numpy as np

from .utils import assert_array_rank, sample_from_dict, assert_string_type
from ._config import get_config

# def _qsample_with_prob_distribs(seq, distrib):

def _qsample_once(seq, qnet, baseline_prob, force_change):
    """Perform one instance of q-sampling.

    NOTE: `seq` is modified in-place

    Parameters
    ----------
    seq : 1d array-like
        Array of values

    qnet : Qnet
        The Qnet that `seq` belongs to 

    baseline_prob : 1d array-like
        Baseline probability for sampling which index

    force_change : bool
        Whether to force the sequence to change when sampling. 
        There is a special cases where the predicted distribution is a single 
        item with 100% probability and the item is exactly the item in the
        corresponding index of `seq`. Then in that case, we are unable to
        force the sequence to change and we simply return the original sequence.


    Returns
    -------
    None
    """

    seq_len = len(seq)

    if seq_len != len(qnet.estimators_):
        message = ('The input sequence length ({}) must match the' 
                  'number of features trained on the qnet ({})')
        warnings.warn(message.format(seq_len, len(qnet.estimators_)))

    # get the index distribution from a distribution
    if baseline_prob is None:
        index = np.random.randint(0, seq_len)
    else:
        index = np.random.choice(
            np.arange(0, seq_len),
            p=baseline_prob)

    # get distribution corresponding to the index
    index_to_non_leaf_nodes = qnet._map_col_to_non_leaf_nodes()
    col_to_item = {i: seq[i] for i in index_to_non_leaf_nodes[index]}
    distrib = qnet.predict_distribution(col_to_item, index)

    if force_change:
        if seq[index] in distrib:
            del distrib[seq[index]]

        if len(distrib) == 0:
            return

        if np.sum(list(distrib.values())) == 0:
            return

    item = sample_from_dict(distrib)

    seq[index] = item

def qsample(seq, qnet, steps, baseline_prob=None, force_change=False):
    """Perform q-sampling for multiple steps.

    Qsampling works as follows: Say you have a sequence and a qnet. Then 
    we randomly pick one of the items in the sequence and then change the
    value of that item based on the prediction of the qnet.

    Parameters
    ----------
    seq : 1d array-like
        Array of values

    qnet : Qnet
        The Qnet that `seq` belongs to 

    steps : int
        Number of steps to run q-sampling

    baseline_prob : 1d array-like
        Baseline probability for sampling which index

    force_change : bool
        Whether to force the sequence to change when sampling. 

    Returns
    -------
    seq : 1d array-like
        q-sampled sequence
    """

    assert_array_rank(seq, 1)
    assert_string_type(seq, 'seq')

    if baseline_prob is not None:
        assert_array_rank(baseline_prob, 1)

    seq = seq.copy()
    for _ in range(steps):
        _qsample_once(
            seq, 
            qnet, 
            baseline_prob,
            force_change=force_change)

    return seq

def targeted_qsample(seq1, seq2, qnet, steps, force_change=False):
    """Perform targeted q-sampling for multiple steps.

    `seq1` is q-sampled towards `seq2`.

    This is similar to `qsample`, except that we perform changes to `seq1`
    to try to approach `seq2`.

    Parameters
    ----------
    seq1 : 1d array-like
        Array of values

    seq2 : 1d array-like
        Array of values. 

    qnet : Qnet
        The Qnet that `seq1` belongs to 

    steps : int
        Number of steps to run q-sampling

    force_change : bool
        Whether to force the sequence to change when sampling. 

    Returns
    -------
    seq : 1d array-like
        q-sampled sequence
    """

    assert_array_rank(seq1, 1)
    assert_array_rank(seq2, 1)
    assert_string_type(seq1, 'seq1')
    assert_string_type(seq2, 'seq2')

    if seq1.shape[0] != seq2.shape[0]:
        raise ValueError('The lengths of the two sequences must be equal!')

    nan_value = get_config()['nan_value']
    seq = seq1.copy()
    seq_len = seq.shape[0]

    for _ in range(steps):

        # only allow different indexes to be sampled
        probs = np.zeros(seq_len, dtype=np.float32)
        probs[seq != seq2] = 1.0
        probs[seq2 == nan_value] = 0.0

        probs /= np.sum(probs)

        _qsample_once(
            seq, 
            qnet, 
            baseline_prob=probs,
            force_change=force_change)

    return seq


import numpy as np
from sklearn.preprocessing import OrdinalEncoder

from ._config import get_config

class OrdinalEncoderWithNaN(OrdinalEncoder):
    """Convert string inputs into ordinal representations with custom NaN value.

    We will encode NaN values with -1.
    """

    def __init__(self, dtype):
        super(OrdinalEncoderWithNaN, self).__init__(dtype=dtype)

    def transform(self, X):

        nan_value = get_config()['nan_value']
        X = super(OrdinalEncoderWithNaN, self).transform(X)
        categories = self.categories_[0]
        if nan_value in categories:
            X[X == list(categories).index(nan_value)] = -1

        return X


from setuptools import setup, Extension, find_packages
from codecs import open
from os import path
import warnings
from setuptools.command.install import install
import atexit
import sys
import os
import platform


class CustomInstall(install):
    def run(self):
        def _post_install():
            def find_module_path():
                for p in sys.path:
                    if os.path.isdir(p) and 'quasinet' in os.listdir(p):
                        return os.path.join(p, 'quasinet')
            install_path = find_module_path()

            # Add your post install code here
            if platform.system() == "Darwin":
                import shutil
                print('copying dcor for mac')
                shutil.copy(install_path+'/bin/macnew_dcor.so',
                            install_path+'/bin/dcor.so')
            else:
                print('no mac exception triggered')
        atexit.register(_post_install)
        install.run(self)




# try:
#     from Cython.Build import cythonize
#     USE_CYTHON = True
# except ImportError:
#     USE_CYTHON = False

package_name = 'quasinet'
example_dir = 'examples/'
example_data_dir = example_dir + 'example_data/'

# ext = '.pyx' if USE_CYTHON else '.c'

# extensions = [
#     Extension("{}".format(package_name), 
#     ["{}/*".format(package_name) + ext])]

# if USE_CYTHON:
#     extensions = cythonize(extensions)
version = {}
with open("version.py") as fp:
    exec(fp.read(), version)

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


setup(
    name=package_name,
    author='zed.uchicago.edu',
    author_email='ishanu@uchicago.edu',
    version = str(version['__version__']),
    packages=find_packages(),
    # package_data={'qnet_trees': ['*/*/*'] },
    scripts=[],
    url='https://github.com/zeroknowledgediscovery/',
    license='LICENSE.txt',
    description='Utitilies for constructing and manipulating models for non-local structural dependencies in genomic sequences',
    keywords=[
        'decision trees', 
        'machine learning', 
        'computational biology'],
    download_url='https://github.com/zeroknowledgediscovery/quasinet/archive/'+str(version['__version__'])+'.tar.gz',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "scikit-learn", 
        "scipy", 
        "numpy", 
        "numba", 
        "pandas",
        "joblib",
        "dill",
        "biopython"],
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Software Development :: Libraries",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6"],
    # data_files=[
    #     ('qnet_example/example1', [
    #         example_dir + "create_qnet.py",
    #         example_dir + "measure_qdistance.py",
    #         example_data_dir + "cchfl_test.csv"
    #         ])
    #     ],
    include_package_data=True,
    cmdclass={'install': CustomInstall},
    # ext_modules=extensions,
    # zip_safe=False
    )



import numpy as np
from numba import njit

@njit(cache=True, nogil=True, fastmath=True)
def kl_divergence(p1, p2):
    """Compute the Kullback–Leibler divergence of discrete probability distributions.

    NOTE: we will not perform error checking in this function because
    this function is used very frequently. The user should check that p1
    and p2 are in fact probability distributions.

    Parameters
    ----------
    p1 : 1d array-like
        probability distribution

    p2 : 1d array-like
        probability distribution

    smooth : float
        amount by which to smooth out the probability distribution of `p2`. This
        is intended to deal with categories with zero probability.

    Returns
    -------
    output : numeric
        kl divergence
    """

    kl_div = (p1 * np.log2(p1 / p2)).sum()

    return kl_div


@njit(cache=True, nogil=True, fastmath=True)
def js_divergence(p1, p2, smooth=0.0001):
    """Compute the Jensen-Shannon of discrete probability distributions.

    Parameters
    ----------
    p1 : 1d array-like
        probability distribution

    p2 : 1d array-like
        probability distribution

    smooth : float
        amount by which to smooth out the probability distribution of `p2`. This
        is intended to deal with categories with zero probability.

    Returns
    -------
    js_div : numeric
        js divergence
    """
    
    p1 = (p1 + smooth) / (1 + smooth)
    p2 = (p2 + smooth) / (1 + smooth)
    p = 0.5 * (p1 + p2)
    js_div = 0.5 * (kl_divergence(p1, p) + kl_divergence(p2, p))

    return js_div